local SEE_WORK_DIST = 14
local KEEP_WORKING_DIST = 20

local WXHorticulture = Class(function(self, inst)
    self.inst = inst
end)

local TOWORK_CANT_TAGS = { "fire", "smolder", "INLIMBO", "NOCLICK", "event_trigger" }
function WXHorticulture:Fertilize()
    local leader = self.inst.components.follower.leader or self.inst.components.entitytracker:GetEntity("sentryward")
    if leader == nil then
        return nil
    end

    local plant = FindEntity(leader, SEE_WORK_DIST, function(ent)
        return ent.components.pickable ~= nil and ent.components.pickable:CanBeFertilized()
    end, nil, TOWORK_CANT_TAGS)
    local fertilizer = self.inst.components.inventory:FindItem(function(item)
        return item.components.fertilizer ~= nil
    end)
    return (plant ~= nil and fertilizer ~= nil) and BufferedAction(self.inst, plant, ACTIONS.FERTILIZE, fertilizer) or nil
end

function WXHorticulture:HarvestFruits()
    local leader = self.inst.components.follower.leader or self.inst.components.entitytracker:GetEntity("sentryward")
    if leader == nil then
        return nil
    end

    local plant = FindEntity(leader, SEE_WORK_DIST, function(ent)
        return ent.components.pickable ~= nil and ent.components.pickable:CanBePicked()
    end, nil, TOWORK_CANT_TAGS)
    return plant ~= nil and BufferedAction(self.inst, plant, ACTIONS.PICK) or nil
end

function WXHorticulture:StoreVeggies()
    local sentryward = self.inst.components.entitytracker:GetEntity("sentryward")
    if sentryward == nil then
        return nil
    end

    local treasurechest = FindEntity(sentryward, SEE_WORK_DIST, function(ent)
        return ent.prefab == "treasurechest" and ent.components.container ~= nil and not ent.components.container:IsFull()
    end)
    local other_product = self.inst.components.inventory:FindItem(function(item)
        return item.components.perishable == nil and item.components.fertilizer == nil
    end)
    if treasurechest ~= nil and other_product ~= nil then
        return BufferedAction(self.inst, treasurechest, ACTIONS.STORE, other_product)
    end

    local saltbox = FindEntity(sentryward, SEE_WORK_DIST, function(ent)
        return ent.prefab == "saltbox" and ent.components.container ~= nil and not ent.components.container:IsFull()
    end)
    local icebox = FindEntity(sentryward, SEE_WORK_DIST, function(ent)
        return ent.prefab == "icebox" and ent.components.container ~= nil and not ent.components.container:IsFull()
    end)

    if saltbox ~= nil then
        local raw_food = self.inst.components.inventory:FindItem(function(item)
            return item.components.edible ~= nil and item.components.edible.foodtype == FOODTYPE.VEGGIE and
                item.components.cookable ~= nil and item:HasTag("cookable")
        end)
        return raw_food ~= nil and BufferedAction(self.inst, saltbox, ACTIONS.STORE, raw_food) or nil
    elseif icebox ~= nil then
        local any_food = self.inst.components.inventory:FindItem(function(item)
            return item.components.edible ~= nil and item.components.perishable ~= nil
        end)
        return any_food ~= nil and BufferedAction(self.inst, icebox, ACTIONS.STORE, any_food) or nil
    end
end

local function FindItemToTakeAction(inst)
    local sentryward = inst.components.entitytracker:GetEntity("sentryward")
    if sentryward == nil then
        return nil
    end

    local target = FindEntity(sentryward, SEE_WORK_DIST, function(chest)
        return chest.prefab == "treasurechest" and chest.components.container ~= nil and
            chest.components.container:FindItem(function(item)
                return item.components.fertilizer ~= nil and
                    inst.components.inventory:FindItem(function(item)
                        return item.components.fertilizer ~= nil
                    end) == nil
            end)
    end)
    local item = nil
    if target ~= nil then
        item = target.components.container:FindItem(function(item)
            return item.components.fertilizer ~= nil
        end)
    end
    return (target ~= nil and item ~= nil) and BufferedAction(inst, target, ACTIONS.LOAD, item) or nil
end

local TOPICKUP_CANT_TAGS = { "fire", "smolder", "INLIMBO", "NOCLICK", "event_trigger", "catchable", "irreplaceable", "heavy", "outofreach" }
function WXHorticulture:FindEntityToPickUpAction()
    local leader = self.inst.components.follower.leader or self.inst.components.entitytracker:GetEntity("sentryward") or self.inst

    local target = FindEntity(leader, SEE_WORK_DIST, function(item)
        return item ~= nil and
        item:IsValid() and
        not item:IsInLimbo() and
        item.entity:IsVisible() and
        item:IsOnValidGround() and
        item.components.inventoryitem ~= nil and
        item.components.inventoryitem.canbepickedup and
        --[[not (item.components.burnable ~= nil and
            (item.components.burnable:IsBurning() or
            item.components.burnable:IsSmoldering())) and]]
        -- Target item is fertilizer, grass or veggie
        ((item.components.fertilizer ~= nil and
        self.inst.components.inventory:FindItem(function(item)
            return item.components.fertilizer ~= nil
        end) == nil) or
        (item.components.edible ~= nil and item.components.edible.foodtype == FOODTYPE.VEGGIE) or
        item.prefab == "cutgrass" or item.prefab == "twigs" or item.prefab == "rock_avocado_fruit")
    end, nil, TOPICKUP_CANT_TAGS)

    return target ~= nil and BufferedAction(self.inst, target, ACTIONS.PICKUP) or FindItemToTakeAction(self.inst) or nil
end

return WXHorticulture